import logging
from asyncio import TaskGroup
from textwrap import indent

from disnake import Embed
from disnake.ext.commands import Cog, group, guild_only

from .utils import can_view, confirm, fetch_exists, fetch_unpacked, test_keyword

log = logging.getLogger("nomen.notifications")
log.setLevel(logging.DEBUG)
log_ = log


class NotifierLogAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return (
            f"[guild: {self.extra["guild"]}, user: {self.extra["user"]}, keyword: {self.extra["keyword"]}] {msg}",
            kwargs,
        )


async def handle_notification(db_updates, ctx, message, keyword, user_id, use_embed):
    """
    Async task to dispatch a notification
    """

    member = await ctx.guild.getch_member(user_id)

    log = NotifierLogAdapter(log_.getChild("notifier"), {"guild": ctx.guild, "user": member, "keyword": keyword})

    log.debug("Handling notification")

    if not await can_view(ctx, member):
        log.debug("Missing permission")
        return

    log.debug("Notifying")
    db_updates.append((ctx.guild.id, keyword, user_id))

    header = f"🔔 `{message.author.display_name}` mentioned `{keyword}` on `{ctx.guild}`:"
    footer = f"\n<t:{int(message.created_at.timestamp())}:R> | [Show]({message.jump_url}) | <#{ctx.channel.id}>"

    if use_embed:
        log.debug("Sending embed")
        embed = Embed(description=message.content + "\n" + footer)
        embed.set_author(
            name=f"{message.author.display_name} ({message.author})",
            icon_url=message.author.display_avatar,
        )

        await member.send(header, embed=embed)
    else:
        log.debug("Sending plain message")
        await member.send("\n".join((header, indent(message.content, "> ", lambda line: True).strip(), footer)))

    log.debug("Sent")


async def handle_triggers(ctx, message):
    """
    Main function that handles message triggers
    """

    log.debug(f"Handling triggers for '{message.content}' from {ctx.author}")

    params = {
        "author": ctx.author.id,
        "channel": ctx.channel.id,
        "guild": ctx.guild.id,
        "content": message.content,
        "is_bot": ctx.author.bot,
    }

    disabled = await fetch_exists(ctx.bot.db, "SELECT * FROM users WHERE user_id=:author AND disabled IS 1", params)

    if disabled:
        log.debug(f"User {ctx.author} ({ctx.author.id}) opted out")
        return

    db_updates = []

    search_query = """
    SELECT keyword, user_id, use_embed
    FROM ( -- Deduplicate notification recipients by user_id
        SELECT keyword, user_id, use_embed, row_number() over (partition by user_id) n
        FROM keywords
        LEFT JOIN users USING (user_id)
        WHERE disabled IS NOT 1                          -- Don't notify users who opted out
        AND (notify_self IS 1 OR user_id IS NOT :author) -- Don't notify author unless wanted
        AND (bots_notify IS 1 OR :is_bot IS NOT 1)       -- Don't notify from bots unless wanted
        AND :author NOT IN (                             -- Don't notify if...
            SELECT target FROM user_ignores                 -- Author is ignored in this guild
            WHERE user_id=user_id AND guild_id=guild_id
            UNION
            SELECT target FROM user_blocks                  -- Author is blocked
            WHERE user_id=user_id
        )
        AND guild_id=:guild AND contains(:content, keyword, regex)
    )
    WHERE n=1
    """

    async with TaskGroup() as tg:
        async with ctx.bot.db.execute(search_query, params) as cur:
            async for keyword, user_id, use_embed in cur:
                log.debug(f"- Creating notification task: '{keyword}' for {user_id}")

                tg.create_task(handle_notification(db_updates, ctx, message, keyword, user_id, use_embed))

    log.debug("Incrementing notification counts")
    await ctx.bot.db.executemany(
        "UPDATE keywords SET count = count + 1 WHERE guild_id=? AND keyword=? AND user_id=?",
        db_updates,
    )
    await ctx.bot.db.commit()


class Notifications(Cog):
    """
    Notifications!
    """

    def __init__(self, bot):
        self.bot = bot

    @Cog.listener()
    async def on_message(self, message):
        if message.author == self.bot.user:
            return

        ctx = await self.bot.get_context(message)
        if ctx.valid or ctx.guild is None:
            return

        await handle_triggers(ctx, message)

    @group(
        aliases=["kw", "notification", "notifications", "notif", "noti"],
        invoke_without_command=True,
    )
    async def keyword(self, ctx):
        """
        Notification keywords
        """
        await ctx.send_help(self.keyword)

    async def _add_keyword(self, ctx, keyword: str, regex: bool):
        log.debug(
            f"Adding {'regex' if regex else 'keyword'}: {keyword} of {ctx.author} ({ctx.author.id}) on {ctx.guild} ({ctx.guild.id})"
        )

        params = {
            "keyword": keyword,
            "guild_id": ctx.guild.id,
            "user_id": ctx.author.id,
            "regex": regex,
        }

        existing = """
        SELECT keyword FROM keywords
        WHERE guild_id=:guild_id
        AND user_id=:user_id
        AND contains(:keyword, keyword, regex)
        """
        redundant = """
        SELECT keyword FROM keywords
        WHERE guild_id=:guild_id
        AND user_id=:user_id
        AND contains(keyword, :keyword, :regex)
        """

        if test_keyword(keyword, regex):
            log.debug("Keyword too common")
            await ctx.send(f"{'Regex' if regex else 'Keyword'} matches a word that is too common")
            return

        conflicts = await fetch_unpacked(ctx.bot.db, existing, params)

        if conflicts:
            log.debug("Keyword conflicts with existing keyword")
            await ctx.send(f"Any instance of `{keyword}` would be matched by existing keywords (check DMs)")
            await ctx.author.send(
                f"Conflicts with keyword `{keyword}`:\n" + "\n".join(f"- `{conflict}`" for conflict in conflicts)
            )
            return

        conflicts = await fetch_unpacked(ctx.bot.db, redundant, params)

        if conflicts:
            log.debug("Keyword renders existing redundant")
            await ctx.send(f"Adding `{keyword}` will cause existing keywords to never match (check DMs)")
            await ctx.author.send(
                f"Keywords redundant from `{keyword}`:\n" + "\n".join(f" - `{conflict}`" for conflict in conflicts)
            )
            return

        log.debug("Keyword valid, adding")

        await ctx.bot.db.execute(
            "INSERT INTO keywords (guild_id, keyword, user_id, regex) VALUES (:guild_id, :keyword, :user_id, :regex)",
            params,
        )
        await ctx.bot.db.commit()
        await ctx.send(f"Added `{keyword}` to your list of keywords")

    @keyword.command()
    @guild_only()
    async def add(self, ctx, keyword):
        """
        Adds a notification keyword

        Use quotes for a keyword with spaces!
        """

        await self._add_keyword(ctx, keyword, False)

    @keyword.command()
    @guild_only()
    async def regex(self, ctx, keyword):
        """
        Adds a notification regex

        Use quotes for a regex with spaces!
        """

        await self._add_keyword(ctx, keyword, True)

    @keyword.command()
    @guild_only()
    async def remove(self, ctx, keyword):
        """
        Removes a keyword or regex

        Must be identical to the keyword or regex you intend on removing
        """

        log.debug(f"Removing keyword: {keyword} of {ctx.author} (ctx.author.id) on {ctx.guild} ({ctx.guild.id})")
        await ctx.bot.db.execute(
            "DELETE FROM keywords WHERE guild_id=? AND keyword=? AND user_id=?",
            (ctx.guild.id, keyword, ctx.author.id),
        )
        await ctx.bot.db.commit()
        await ctx.send(f"Removed `{keyword}` from your list of keywords")

    @keyword.command()
    @guild_only()
    async def list(self, ctx):
        """
        Lists keywords and regexes, with trigger count
        """

        log.debug(f"Listing keywords: {ctx.author} ({ctx.author.id}) on {ctx.guild} ({ctx.guild.id})")

        async with ctx.bot.db.execute(
            "SELECT keyword, count, regex FROM keywords WHERE guild_id=? AND user_id=?",
            (ctx.guild.id, ctx.author.id),
        ) as cur:
            keywords = await cur.fetchall()

        embed = Embed(
            description="\n".join(
                f"{'regex '*regex}`{keyword}`: notified `{count}` times" for keyword, count, regex in keywords
            )
        )
        embed.set_author(name=f"Your keywords on {ctx.guild}:", icon_url=ctx.guild.icon)

        await ctx.author.send(embed=embed)
        await ctx.send("Please check your direct messages")

    @keyword.command()
    @guild_only()
    async def clear(self, ctx):
        """
        Clears keywords and regexes for this server
        """

        log.debug(f"Clearing keywords: {ctx.author} ({ctx.author.id}) on {ctx.guild} ({ctx.guild.id})")

        to_clear = await confirm(
            ctx, "Are you sure you want to clear your keywords on {ctx.guild}? (Reply with yes/no)", True
        )

        if to_clear is None:
            return await ctx.send("Timed out or received an invalid response", delete_after=10)
        elif not to_clear:
            return await ctx.send("Canceled", delete_after=10)

        await ctx.send("Clearing keywords")

        await ctx.bot.db.execute(
            "DELETE FROM KEYWORDS WHERE guild_id=? AND user_id=?",
            (ctx.guild.id, ctx.author.id),
        )
        await ctx.bot.db.commit()

    @keyword.command()
    @guild_only()
    async def pause(self, ctx):
        pass
        # TODO: Pause guild notifications

    @keyword.group(invoke_without_command=True)
    @guild_only()
    async def ignore(self, ctx, channel):
        pass
        # TODO: Ignore channels and users

    @ignore.command()
    @guild_only()
    async def active(self, ctx):
        pass
        # TODO: Ignore active channel
